#define ALLOW_ACCESS_TO_INTERNALS_OF_SCROLLBARS
#define ALLOW_ACCESS_TO_INTERNALS_OF_CONTROLS
#define ALLOW_ACCESS_TO_INTERNALS_OF_WINDOWS
#define ALLOW_ACCESS_TO_INTERNALS_OF_FIELDS
#define ALLOW_ACCESS_TO_INTERNALS_OF_TABLES
#define ALLOW_ACCESS_TO_INTERNALS_OF_LISTS
#define ALLOW_ACCESS_TO_INTERNALS_OF_FORMS
#include <PalmOS.h>

struct WinGlobalsPart {	// @0x18c
	WindowType* someImportantWindows[7];
};

#define WIN_IDX_LIST_HEAD		3

struct FrmGlobalsPart {	// @0x1f0
	FormType *curForm;
};

struct FormActiveStateInsides {		//22 bytes max
	UInt32 magix;
	UInt16 insPtrOn	:	 1;
	
	Int16 insPtX, insPtY;
	FormPtr prevCurForm;
	WinHandle prevDrawWin, prevActiveWin;	//only if no active form
};

#define SAVE_STATE_MAGIX			'fSS!'

static Err pFrmActiveState(struct FormActiveStateInsides *ss, Boolean save)
{
	
	if (save) {
	
		MemSet(ss, sizeof(*ss), 0);
		ss->magix = SAVE_STATE_MAGIX;
		ss->insPtrOn = InsPtEnabled();
		InsPtGetLocation(&ss->insPtX, &ss->insPtY);
		if (!(ss->prevCurForm = FrmGetActiveForm())) {
	
			ss->prevDrawWin = WinGetDrawWindow();
			ss->prevActiveWin = WinGetActiveWindow();
		}
	}
	else if (ss->magix != SAVE_STATE_MAGIX)
		return sysErrParamErr;
	else {
		
		InsPtSetLocation(ss->insPtX, ss->insPtY);
		if (ss->prevCurForm)
			FrmSetActiveForm(ss->prevCurForm);
		else {
			
			WinSetDrawWindow(ss->prevDrawWin);
			WinSetActiveWindow(ss->prevActiveWin);
		}
		InsPtEnable(ss->insPtrOn);
	}
	
	return errNone;
}


static void* prvFormAdjustPointer(void *srcPtr, UInt8 *movedFrom, UInt8 *movedFromEnd, Int32 movedBy)
{
	UInt8 *src = srcPtr;
	
	if (src >= movedFrom && src < movedFromEnd)
		src += movedBy;
	
	return src;
}

static void prvFormPieceMoved(FormType *fp, UInt8 *movedFrom, UInt16 moveLen, Int32 movedBy)
{
	UInt8 *movedFromEnd = movedFrom + moveLen;
	UInt16 i;

	//special case for empty forms (ask me how i figured this out, ugh!)
	if (!fp->numObjects && (UInt8*)fp->objects == movedFromEnd)
		movedFromEnd++;
	
	
	fp->objects = prvFormAdjustPointer(fp->objects, movedFrom, movedFromEnd, movedBy);
	for (i = 0; i < fp->numObjects; i++) {
		
		void *obj = fp->objects[i].object.ptr = prvFormAdjustPointer(fp->objects[i].object.ptr, movedFrom, movedFromEnd, movedBy);
		FormTitleType *title = (FormTitleType*)obj;
		FormLabelType *lbl = (FormLabelType*)obj;
		ControlType *ctl = (ControlType*)obj;
		FieldType *field = (FieldType*)obj;
		TableType *tbl = (TableType*)obj;
		ListType *list = (ListType*)obj;
		
		
		switch (fp->objects[i].objectType) {
		
			//Not all objects need to be moved, only those that can be set by ResLoadForm().
			// Other pointers cannot point into the memory chunk owned by the form.
			default:
				break;
		
			case frmFieldObj:
				
				field->lines = prvFormAdjustPointer(field->lines, movedFrom, movedFromEnd, movedBy);
				break;
			
			case frmControlObj:
				if (!ctl->attr.graphical)
					ctl->text = prvFormAdjustPointer(ctl->text, movedFrom, movedFromEnd, movedBy);
				break;
			
			case frmListObj:
				if (list->itemsText) {
					
					Int16 idx;
					
					list->itemsText = prvFormAdjustPointer(list->itemsText, movedFrom, movedFromEnd, movedBy);
					
					for (idx = 0; idx < list->numItems; idx++)
						list->itemsText[idx] = prvFormAdjustPointer(list->itemsText[idx], movedFrom, movedFromEnd, movedBy);
				}
				break;
			
			case frmTableObj:
				tbl->columnAttrs = prvFormAdjustPointer(tbl->columnAttrs, movedFrom, movedFromEnd, movedBy);
				tbl->rowAttrs = prvFormAdjustPointer(tbl->rowAttrs, movedFrom, movedFromEnd, movedBy);
				tbl->items = prvFormAdjustPointer(tbl->items, movedFrom, movedFromEnd, movedBy);
				break;
			
			case frmLabelObj:
				lbl->text = prvFormAdjustPointer(lbl->text, movedFrom, movedFromEnd, movedBy);
				break;
			
			case frmTitleObj:
				title->text = prvFormAdjustPointer(title->text, movedFrom, movedFromEnd, movedBy);
				break;
		}
	}
}

static void prvFormPieceDoMove(FormType *fp, UInt8 *moveDst, UInt8 *moveSrc, UInt16 moveLen)
{
	MemMove(moveDst, moveSrc, moveLen);
	prvFormPieceMoved(fp, moveSrc, moveLen, moveDst - moveSrc);
}

static Err pFrmRemoveObject(FormType **fPP, UInt16 objIndex)
{
	UInt16 origSize, newSize;
	UInt8 *origStart;
	FormType *fp;


	fp = *fPP;
	origStart = (UInt8*)fp;
	newSize = origSize = MemPtrSize(fp);
	
	//move object data
	if (objIndex < fp->numObjects - 1) {
	
		prvFormPieceDoMove(fp, fp->objects[objIndex + 1].object.ptr, fp->objects[objIndex].object.ptr, origStart + origSize - (UInt8*)fp->objects[objIndex + 1].object.ptr);
		newSize -= (UInt8*)fp->objects[objIndex + 1].object.ptr - (UInt8*)fp->objects[objIndex].object.ptr;
	}
	
	//move to remove space in the object list
	prvFormPieceDoMove(fp, (UInt8*)(fp->objects + objIndex), (UInt8*)(fp->objects + objIndex + 1), origStart + newSize - (UInt8*)(fp->objects + objIndex + 1));
	newSize -= sizeof(FormObjListType);
	
	//accounting
	fp->numObjects--;
	
	//resize
	MemPtrResize(fp, newSize);
	
	return errNone;
}

static Err pFrmAddSpaceForObject(FormType **fPP, void **objPP, FormObjectKind objectKind, UInt16 objSize)
{
	struct FrmGlobalsPart *g = (struct FrmGlobalsPart*)0x1f0;
	UInt8 *oldStart, *newStart;
	UInt16 origSize, newSize;
	MemHandle fH;
	FormType *fp;
	Err e;
	
	
	fH = MemPtrRecoverHandle(oldStart = (UInt8*)*fPP);
	origSize = (MemHandleSize(fH) + 1) &~ 1;
	newSize = origSize + sizeof(FormObjListType) + objSize;
	
	MemHandleUnlock(fH);
	e = MemHandleResize(fH, newSize);
	if (e != errNone) {
		MemHandleLock(fH);
		return e;
	}
	*fPP = fp = MemHandleLock(fH);
	newStart = (UInt8*)fp;
	
	//if the form was moved due to the resize, relocate all the pointers
	if (newStart != oldStart) {
		
		prvFormPieceMoved(fp, oldStart, origSize, newStart - oldStart);
		
		if (g->curForm == (void*)oldStart)
			g->curForm = (void*)newStart;
		
		WinMoveWindowAddr((WindowType*)oldStart, (WindowType *)newStart);
	}
	
	//and now a partial move (data moves, object list stays) to make space for a ne wobject list entry
	prvFormPieceDoMove(fp, (UInt8*)(fp->objects + fp->numObjects + 1), (UInt8*)(fp->objects + fp->numObjects), (newStart + origSize) - (UInt8*)(fp->objects + fp->numObjects));
	
	
	//make space for data and zero it
	MemSet(*objPP = newStart + newSize - objSize, objSize, 0);
	
	//init the new space
	fp->objects[fp->numObjects].objectType = objectKind;
	fp->objects[fp->numObjects].object.ptr = *objPP;
	fp->numObjects++;
	
	return errNone;
}

static FormType* pFrmNewForm(UInt16 formID, const char *titleStrP, Coord x, Coord y, Coord width, Coord height, Boolean modal, UInt16 defaultButton, UInt16 helpRscID, UInt16 menuRscID)
{
	UInt16 memSize = sizeof(FormType);
	MemHandle fH;
	FormType *fp;
	
	
	if (titleStrP)
		memSize += sizeof(FormObjListType) +  sizeof(FormTitleType) + StrLen(titleStrP) + 1;
	
	fH = MemHandleNew(memSize);
	if (!fH)
		return NULL;
	fp = MemHandleLock(fH);

	MemSet(fp, memSize, 0);

	if (!modal)
		fp->window.frameType.word = noFrame;
	else {
	
		fp->window.windowFlags.modal = modal;
		fp->window.frameType.word = dialogFrame;
	}
	
	fp->window.windowBounds.topLeft.x = x;
	fp->window.windowBounds.topLeft.y = y;
	fp->window.windowBounds.extent.x = width;
	fp->window.windowBounds.extent.y = height;
	fp->window.windowFlags.dialog = true;
	fp->window.windowFlags.focusable = true;
	fp->attr.usable = true;
	fp->attr.saveBehind = true;
	fp->attr.attnIndicator = false;
	fp->formId = formID;
	fp->helpRscId = helpRscID;
	fp->menuRscId = menuRscID;
	fp->defaultButton = defaultButton;
	fp->focus = noFocus;
	fp->objects = (FormObjListType*)(fp + 1);
	
	if (titleStrP) {
		
		FormTitleType *title = (FormTitleType*)(fp->objects + 1);
		
		fp->numObjects = 1;
		fp->objects[0].objectType = frmTitleObj;
		fp->objects[0].object.ptr = title;
		
		StrCopy(title->text = (char*)(title + 1), titleStrP);
	}
	
	WinInitializeWindow((WinHandle)&fp->window);
	WinAddWindow((WinHandle)&fp->window);
	
	return fp;
}

static void pWinMoveWindowAddr(WindowType* oldPtr, WindowType *newPtr)
{
	struct WinGlobalsPart *g = (struct WinGlobalsPart*)0x18c;
	WindowType *iter;
	UInt8 i;
	
	//find it in the window list
	for (iter = g->someImportantWindows[WIN_IDX_LIST_HEAD]; iter && iter->nextWindow != oldPtr; iter = iter->nextWindow);
	if (iter)
		iter->nextWindow = newPtr;
	
	//see if this was a tracked window, and repoint as needed
	for (i = 0; i < sizeof(g->someImportantWindows) / sizeof(*g->someImportantWindows); i++) {
		if (g->someImportantWindows[i] == oldPtr)
			g->someImportantWindows[i] = newPtr;
	}
}

static FieldType* pFldNewField(FormType **fPP, UInt16 id, Coord x, Coord y, Coord width, Coord height, FontID font, UInt32 maxChars, Boolean editable, Boolean underlined, Boolean singleLine, Boolean dynamicSize, JustificationType justification, Boolean autoShift, Boolean hasScrollBar, Boolean numeric)
{
	FieldType *fld;
	
	if (errNone != FrmAddSpaceForObject(fPP, (void**)&fld, frmFieldObj, sizeof(FieldType)))
		return NULL;
	
	fld->id = id;
	fld->rect.topLeft.x = x;
	fld->rect.topLeft.y = y;
	fld->rect.extent.x = width;
	fld->rect.extent.y = height;
	fld->attr.usable = true;
	fld->attr.editable = editable;
	if (underlined)
		fld->attr.underlined = grayUnderline;
	fld->attr.singleLine = singleLine;
	fld->attr.dynamicSize = dynamicSize;
	fld->attr.autoShift = autoShift;
	fld->attr.hasScrollBar = hasScrollBar;
	fld->attr.numeric = numeric;
	fld->attr.justification = justification;
	fld->attr.insPtVisible = true;	
	fld->maxChars = maxChars;
	fld->fontID = font;

	return fld;
}

static ControlType* pCtlNewControl(FormType **fPP, UInt16 ID, ControlStyleType style, const char *textP, Coord x, Coord y, Coord width, Coord height, FontID font, UInt8 group, Boolean leftAnchor)
{
	UInt16 objSz = (style == sliderCtl || style == feedbackSliderCtl) ? sizeof(SliderControlType) : sizeof(ControlType);
	ControlType *ctl;

	if (textP)
		objSz += StrLen(textP) + 1;

	if (errNone != FrmAddSpaceForObject(fPP, (void**)&ctl, frmControlObj, objSz))
		return NULL;
	
	ctl->id = ID;
	ctl->style = style;
	ctl->attr.frame = (style == buttonCtl || style == pushButtonCtl) ? standardButtonFrame : noButtonFrame;
	ctl->bounds.topLeft.x = x;
	ctl->bounds.topLeft.y = y;
	ctl->attr.usable = true;
	ctl->attr.enabled = true;
	ctl->attr.leftAnchor = leftAnchor;
	ctl->font = font;
	ctl->group = group;

	//sort out width
	if (width)
		ctl->bounds.extent.x = width;
	else if (textP) {	//calc it
	
		CtlSetLabel (ctl, textP);		//temporarily points contorl's internal text pointer to passed-in pointer. long term this woudl be bad, but for now it is not, we'll sort it out later
		
		//re-re-check
		if (!ctl->bounds.extent.x)
			ctl->bounds.extent.x = FntCharsWidth(textP, StrLen(textP)) + 9;
	}
	
	//sort out height
	if (height)
		ctl->bounds.extent.y = height;
	else {				//calc it
		
		FontID prevFont;
		
		prevFont = FntSetFont(font);
		ctl->bounds.extent.y = FntLineHeight() + 1;
		FntSetFont(prevFont);
	}

	//sort out label
	if (textP)
		StrCopy(ctl->text = (char*)(ctl + 1), textP);
	
	return ctl;
}

static Err pLstNewList(FormType **fPP, UInt16 id, Coord x, Coord y, Coord width, Coord height, FontID font, Int16 visibleItems, Int16 triggerId)
{
	FontID prevFont;
	ListType *list;
	Err e;
	
	e = FrmAddSpaceForObject(fPP, (void**)&list, frmListObj, sizeof(ListType));
	if (e != errNone)
		return e;

	prevFont = FntSetFont(font);
	list->id = id;
	list->bounds.topLeft.x = x;
	list->bounds.topLeft.y = y;
	list->bounds.extent.x = width;
	list->bounds.extent.y = FntLineHeight() * visibleItems;
	list->font = font;
	FntSetFont(prevFont);

	if (!triggerId)
		list->attr.usable = true;
	else {
	
		FormPopupType *popup;
		
		e = FrmAddSpaceForObject(fPP, (void**)&popup, frmPopupObj, sizeof(FormPopupType));
		if (e != errNone)
			return e;
		
		popup->controlID = triggerId;
		popup->listID = id;
	}

	return errNone;
}

static FormLabelType* pFrmNewLabel(FormType **fPP, UInt16 ID, const char *textP, Coord x, Coord y, FontID font)
{
	FormLabelType *lbl;
	Err e;

	if (errNone != FrmAddSpaceForObject(fPP, (void**)&lbl, frmLabelObj, sizeof(FormLabelType) + (textP ? StrLen(textP) + 1 : 0)))
		return NULL;
	
	lbl->id = ID;
	lbl->pos.x = x;
	lbl->pos.y = y;
	lbl->attr.usable = true;
	lbl->fontID = font;

	if (textP)
		StrCopy(lbl->text = (char*)(lbl + 1), textP);
	
	return lbl;
}

static FormBitmapType* pFrmNewBitmap (FormType **fPP, UInt16 ID, UInt16 rscID, Coord x, Coord y)
{
	FormBitmapType *bmp;
	
	if (errNone != FrmAddSpaceForObject(fPP, (void**)&bmp, frmBitmapObj, sizeof(FormBitmapType)))
		return NULL;
	
	bmp->pos.x = x;
	bmp->pos.y = y;
	bmp->attr.usable = true;
	bmp->rscID = rscID;
	
	return bmp;
}

static FormGadgetType* pFrmNewGadget(FormType **fPP, UInt16 id, Coord x, Coord y, Coord width, Coord height)
{
	FormGadgetType *gdgt;
	
	if (errNone != FrmAddSpaceForObject(fPP, (void**)&gdgt, frmGadgetObj, sizeof(FormGadgetType)))
		return NULL;
	
	gdgt->id = id;
	gdgt->rect.topLeft.x = x;
	gdgt->rect.topLeft.y = y;
	gdgt->rect.extent.x = width;
	gdgt->rect.extent.y = height;
	gdgt->attr.usable = true;
	gdgt->attr.extended = true;
	
	return gdgt;
}

void __Startup__(void)
{
	UInt32 ver;
	
	if (errNone == FtrGet(sysFtrCreator, sysFtrNumROMVersion, &ver) && ver < sysMakeROMVersion(3, 0, 0, 0, 0)) {
	
		SysSetTrapAddress(sysTrapWinMoveWindowAddr, pWinMoveWindowAddr);
		SysSetTrapAddress(sysTrapFrmAddSpaceForObject, pFrmAddSpaceForObject);
		SysSetTrapAddress(sysTrapFrmNewForm, pFrmNewForm);
		SysSetTrapAddress(sysTrapCtlNewControl, pCtlNewControl);
		SysSetTrapAddress(sysTrapFldNewField, pFldNewField);
		SysSetTrapAddress(sysTrapLstNewList, pLstNewList);
		SysSetTrapAddress(sysTrapFrmNewLabel, pFrmNewLabel);
		SysSetTrapAddress(sysTrapFrmNewBitmap, pFrmNewBitmap);
		SysSetTrapAddress(sysTrapFrmNewGadget, pFrmNewGadget);
		SysSetTrapAddress(sysTrapFrmRemoveObject, pFrmRemoveObject);
		SysSetTrapAddress(sysTrapFrmActiveState, pFrmActiveState);
		
		MemHandleLock(DmGet1Resource('extn', 6));
	}
}
