#include <PalmOS.h>

#define DESIRED_NUM_TRAPS			0x3D9		//up to OS 3.5 level + a few from 4.0

#define OWNER_ID_FTR_ALLOCATED_CHUNKS	3

struct PalmOsGlobals {		//at 0x02 to make gcc happy
	UInt8 whatever1[0x120];
	void** trapTable;			//0x122
	UInt8 whatever2[0x017];
	UInt8 numTrapsModified;		//0x13D
	UInt16 numTraps;			//0x13E
};
#define GLOBALS_INITIALIZER			((struct PalmOsGlobals*)2)

static Err pSysSetTrapAddress(UInt16 trapNum, void *procP)
{
	struct PalmOsGlobals *pg = GLOBALS_INITIALIZER;

	if (trapNum < sysTrapBase)
		return sysErrParamErr;
	
	trapNum -= sysTrapBase;
	
	if (trapNum >= pg->numTraps)
		return sysErrParamErr;
	
	DmWrite(pg->trapTable, (char*)&pg->trapTable[trapNum] - (char*)pg->trapTable, &procP, 4);
	pg->numTrapsModified++;
	
	return errNone;
}

static void* pSysGetTrapAddress(UInt16 trapNum)
{
	struct PalmOsGlobals *pg = GLOBALS_INITIALIZER;
	
	if (trapNum < sysTrapBase)
		trapNum = sysTrapSysUnimplemented;
	
	trapNum -= sysTrapBase;
	
	if (trapNum >= pg->numTraps)
		trapNum = sysTrapSysUnimplemented - sysTrapBase;
	
	return pg->trapTable[trapNum];
}

void __Startup__(void)
{
	struct PalmOsGlobals *pg = GLOBALS_INITIALIZER;
	
	if (pg->numTraps < DESIRED_NUM_TRAPS) {
		
		UInt16 i, numTraps = pg->numTraps, numHeaps;
		void **newTab, *unimpl;
		MemHandle newTabH;

		newTabH = MemChunkNew(1, DESIRED_NUM_TRAPS * 4, memNewChunkFlagAtEnd | OWNER_ID_FTR_ALLOCATED_CHUNKS);
		if (!newTabH)
			SysFatalAlert("ERR: new tab");
		newTab = MemHandleLock(newTabH);
		
		//copy existing traps
		DmWrite(newTab, 0, pg->trapTable, sizeof(UInt32) * numTraps);
		
		//set all new traps to be unimplemented
		for (i = numTraps; i < DESIRED_NUM_TRAPS; i++)
			DmWrite(newTab, (char*)&newTab[i] - (char*)newTab, &newTab[sysTrapSysUnimplemented - sysTrapBase], 4);
		
		//re-point the table
		pg->trapTable = newTab;
		
		//tell the parts of the OS that care (spoiler alert: none do)
		pg->numTraps = DESIRED_NUM_TRAPS;
		
		//need to call directly to avoid old funcs being used and faulting
		pSysSetTrapAddress(sysTrapSysSetTrapAddress, &pSysSetTrapAddress);
		pSysSetTrapAddress(sysTrapSysGetTrapAddress, &pSysGetTrapAddress);
		
		MemHandleLock(DmGet1Resource('extn', 1));	//we have patches in this code resource
	}
}



