#include <ExpansionMgr.h>
#include <SlotDrvrLib.h>
#include <VFSMgr.h>
#include <PalmOS.h>
#include "iface.h"
#include "proto.h"

static const char mHwNames[][8] = {"TEST", "SASD v1"};
static const UInt32 mSpeedLimits[] = {57600, 115200, 230400, 300000, 400000, 550000, 700000};	//must preceisely match list in UI

#ifndef expCapabilityHidden
#define expCapabilityHidden		8
#endif

//SonyStyle
	struct CardPartitionLayoutV1 {			//allocated by us with owner of zero. cfeed by caller
		UInt32 version;						// 0x00 - the value 1 is is stored here
		UInt32 partStartMinus1;				// 0x04 - if part is to start at sector 0x19, then 0x18 will be stored here
	};

	struct CardMetricsTypeV1 {				//used by expansion manager v1 (sony 3.5)
		UInt32 version;						// 0x00 - the value 1 is is stored here
		UInt32 totalSectors;				// 0x04
		UInt16 bytesPerSector;				// 0x08
		UInt8 sectorsPerBlock;				// 0x0a
		UInt8 rfu;							// 0x0b - the value 0 is written here
		UInt16 sectorsPerHead;				// 0x0c
		UInt16 headsPerCylinder;			// 0x0e
		UInt32 unknown_1;					// 0x10 - the value 1 is is stored here
		struct CardPartitionLayoutV1 *pl;	// 0x14 - allocated by the driver and freed by called. 8 bytes
	};

	union Metrics {
		CardMetricsType v2;
		struct CardMetricsTypeV1 v1;
	};

static Boolean isSonyEarlyVFS(void)	//PalmOS 3.5 Sony devices have an early implementation of ExpMgr/VFSMgr that differs in a few ways that we care about. This funcs tells us if this is such a device
{
	UInt32 expnVer;
	
	return errNone == FtrGet(sysFileCExpansionMgr, expFtrIDVersion, &expnVer) && expnVer == 1;
}

static void uiPrvSetBaudrateLimit(UInt32 rate)
{
	//version "-1" will cause pref getting to fail, as desired
	PrefSetAppPreferencesV10(SASD_PREFS_CREATOR, rate ? 0 : -1, &rate, sizeof(rate));
}

static void uiPrvLargeNumberToString(char *str, char separator, UInt32 value, UInt32 printWidthMinAsIfValue)
{
	UInt16 minLen, haveLen, i, j, outPos, inPos, nFullGroups;
	char x[16];
	
	
	//make sure printWidthMinAsIfValue > value
	if (printWidthMinAsIfValue < value)
		printWidthMinAsIfValue = value;
	
	//sort out lengths
	StrIToA(x, printWidthMinAsIfValue);
	minLen = StrLen(x);
	StrIToA(x, value);
	haveLen = StrLen(x);
	
	//pad front
	for (i = 0; i < minLen - haveLen; i++)
		x[i] = '0';
	StrIToA(x + i, value);
	haveLen = minLen;
	
	nFullGroups = haveLen / 3;
	haveLen %= 3;
	
	inPos = 0;
	outPos = 0;

	for (i = 0; i < haveLen; i++)
		str[outPos++] = x[inPos++];

	for (i = 0; i < nFullGroups; i++) {
		
		if (outPos)
			str[outPos++] = separator;
		
		for (j = 0; j < 3; j++)
			str[outPos++] = x[inPos++];
	}
	str[outPos] = 0;
}

static void uiPrvSetControl(FormPtr fp, UInt16 id, Boolean on)
{
	CtlSetValue(FrmGetObjectPtr(fp, FrmGetObjectIndex(fp, id)), on);
}

static void uiPrvCardInfo(UInt16 slotLibRefNum, UInt16 slotRefNum)
{
	if (errNone != ExpCardPresent(slotRefNum))
		FrmCustomAlert(1000, "No card present", " ", " ");
	else {
		
		FormPtr fp = FrmInitForm(1002);
		ExpCardInfoType ci = {0, };
		union Metrics me;
		UInt32 flags;
		UInt16 i, sz;
		char x[32];
		Err e;
		
		if (errNone == ExpCardInfo(slotRefNum, &ci)) {
			
			FrmCopyLabel(fp, 1000, ci.manufacturerStr);
			FrmCopyLabel(fp, 1001, ci.productStr);
			FrmCopyLabel(fp, 1002, ci.deviceClassStr);
			FrmCopyLabel(fp, 1003, ci.deviceUniqueIDStr);
			
			uiPrvSetControl(fp, 2000, !!(ci.capabilityFlags & expCapabilityReadOnly));
			uiPrvSetControl(fp, 2000, !!(ci.capabilityFlags & expCapabilityHidden));
		}
		
		if (errNone == SlotCardMetrics(slotLibRefNum, slotRefNum, (CardMetricsPtr)&me)) {
			
			UInt32 nSec;
			
			if (isSonyEarlyVFS()) {
				nSec = me.v1.totalSectors;
				MemPtrFree(me.v1.pl);
			}
			else
				nSec = me.v2.totalSectors;
			
			uiPrvLargeNumberToString(x, ',', nSec / 2, 0);
			StrCat(x, " KB");
			
			FrmCopyLabel(fp, 1005, x);
		}
		
		sz = sizeof(flags);
		if (errNone == SlotCustomControl(slotLibRefNum, SASD_API_CREATOR, SASD_GET_FLAGS, &flags, &sz) && sz == sizeof(flags)) {
		
			uiPrvSetControl(fp, 3000, !!(flags & SD_FLAG_IS_MMC));
			uiPrvSetControl(fp, 3001, !!(flags & SD_FLAG_IS_HIGH_CAP));
			uiPrvSetControl(fp, 3002, !(flags & SD_FLAG_IS_MMC));
			uiPrvSetControl(fp, 3003, !!(flags & SD_FLAG_READ_ONLY));
			uiPrvSetControl(fp, 3004, !!(flags & SD_FLAG_HAS_DISCARD));
			uiPrvSetControl(fp, 3005, !!(flags & SD_FLAG_HAS_CMD_23));
			uiPrvSetControl(fp, 3006, !!(flags & SD_FLAG_IS_NEW_VER));
		}
		while (FrmDoDialog(fp) == 8000) {
			
				
			VFSSlotMountParamType mpt = {.vfsMountParam = {.mountClass = vfsMountClass_SlotDriver, }, .slotLibRefNum = slotLibRefNum, .slotRefNum = slotRefNum, };
	
			e = VFSVolumeFormat(0, 0, &mpt.vfsMountParam);
			if (e == errNone)
				FrmCustomAlert(1004, "Card formatted successfully", " ", " ");
			else {
				
				StrCopy(x, "Format error 0x");
				StrIToH(x + StrLen(x), e);
				FrmCustomAlert(1000, x, " ", " ");
			}
		}
		FrmDeleteForm(fp);
	}
}

static void uiPrvAbout(void)
{
	FormPtr fp = FrmInitForm(1001);
	FrmDoDialog(fp);
	FrmDeleteForm(fp);
}

static void uiPrvFirmwareUpdate(UInt16 slotLibRefNum)
{
	UInt16 sz = 0;
	
	if (FrmCustomAlert(1002, "This will erase the firmware in your SASD device in preparation for flashing a new one. The device will not work until you flash it. Are you sure you want to proceed?", " ", " ")) {
		
		if (errNone == SlotCustomControl(slotLibRefNum, SASD_API_CREATOR, SASD_FW_UPDATE_START, &sz, &sz) && !sz)
			FrmAlert(1003);
		else
			FrmCustomAlert(1000, "Failed to kick device into update mode", " ", " ");
	}							
}

static void ui(UInt16 slotLibRefNum, UInt16 slotRefNum)
{
	struct SdVersionInfo svi;
	struct SdBatteryData sbi;
	ControlType *pop;
	ListType *list;
	UInt32 ifaceSpeed;
	EventType evt;
	DmOpenRef db;
	MemHandle mh;
	UInt16 i, sz;
	char x[32];
	FormPtr fp;
	Err e;
	
	
	fp = FrmInitForm(1000);
	
	//driver ver
	db = DmOpenDatabase(0, DmFindDatabase(0, "SASD"), dmModeReadOnly);
	if (db) {
		mh = DmGet1Resource(verRsc, appVersionAlternateID);
		if (mh) {
			
			FrmCopyLabel(fp, 1000, MemHandleLock(mh));
			MemHandleUnlock(mh);
			DmReleaseResource(mh);
		}
		DmCloseDatabase(db);
	}
	
	//device versions
	sz = sizeof(svi);
	if (errNone == SlotCustomControl(slotLibRefNum, SASD_API_CREATOR, SASD_GET_VERS, &svi, &sz) && sz == sizeof(svi)) {
		
		if (svi.hwVers < sizeof(mHwNames) / sizeof(*mHwNames))
			FrmCopyLabel(fp, 1001, mHwNames[svi.hwVers]);
		else {
			StrCopy(x, "UNKNOWN: ");
			StrIToA(x + StrLen(x), svi.hwVers);
			FrmCopyLabel(fp, 1001, x);
		}
		
		StrIToA(x, svi.fwVers[0]);
		StrCat(x, ".");
		StrIToA(x + StrLen(x), svi.fwVers[1]);
		StrCat(x, ".");
		StrIToA(x + StrLen(x), svi.fwVers[2]);
		FrmCopyLabel(fp, 1002, x);
	}
	
	//battery
	sz = sizeof(sbi);
	if (errNone == SlotCustomControl(slotLibRefNum, SASD_API_CREATOR, SASD_BATTERY_LVL, &sbi, &sz) && sz == sizeof(sbi)) {
		
		uiPrvLargeNumberToString(x, '.', sbi.voltage, 1000);
		StrCat(x, "V (");
		StrIToA(x + StrLen(x), sbi.percent);
		StrCat(x, "%)");
		FrmCopyLabel(fp, 1003, x);
	}
	
	//iface speed
	sz = sizeof(ifaceSpeed);
	if (errNone == SlotCustomControl(slotLibRefNum, SASD_API_CREATOR, SASD_GET_IFACE_SPEED, &ifaceSpeed, &sz) && sz == sizeof(ifaceSpeed) && ifaceSpeed) {
		
		uiPrvLargeNumberToString(x, ',', ifaceSpeed, 0);
		StrCat(x, " bps");
		FrmCopyLabel(fp, 1004, x);
	}
	
	//current limit
	pop = FrmGetObjectPtr(fp, FrmGetObjectIndex(fp, 1005));
	list = FrmGetObjectPtr(fp, FrmGetObjectIndex(fp, 3000));
	
	if (!PrefGetAppPreferencesV10(SASD_PREFS_CREATOR, 0, &ifaceSpeed, sizeof(ifaceSpeed)))
		ifaceSpeed = 230400; //this should be kept in sync with "maxSpeed.c" in driver src
	
	//find exact entry and mark it
	for (i = 0; i < sizeof(mSpeedLimits) / sizeof(*mSpeedLimits); i++) {
		
		if (mSpeedLimits[i] == ifaceSpeed) {
			
			CtlSetLabel(pop, LstGetSelectionText(list, i));
			LstSetSelection(list, i);
			break;
		}
	}
	//else write the string and mark nothing
	if (i == sizeof(mSpeedLimits) / sizeof(*mSpeedLimits)) {	//not found?
		
		uiPrvLargeNumberToString(x, ',', ifaceSpeed, 0);
		StrCat(x, " bps");
		LstSetSelection(list, -1);
		FrmCopyLabel(fp, 1004, x);
	}
	
	FrmSetActiveForm(fp);
	FrmDrawForm(fp);
	
	do {
		
		EvtGetEvent(&evt, -1);
		
		switch (evt.eType) {
			
			case ctlSelectEvent: switch (evt.data.ctlSelect.controlID) {
					case 2000:
						uiPrvCardInfo(slotLibRefNum, slotRefNum);
						continue;
					
					case 2001:
						uiPrvFirmwareUpdate(slotLibRefNum);
						continue;
					
					case 2002:
						FrmHelp(1000);
						continue;
					
					case 2003:
						uiPrvAbout();
						continue;
				}
				break;
			
			case frmOpenEvent:
			case frmGotoEvent:
			case frmUpdateEvent:
				FrmDrawForm(fp);
				break;
			
			case popSelectEvent:
				uiPrvSetBaudrateLimit(mSpeedLimits[evt.data.popSelect.selection]);
				break;
		}
		
		if (SysHandleEvent(&evt))
			continue;
		
		if (MenuHandleEvent(NULL, &evt, &e))
			continue;
		
		FrmHandleEvent(fp, &evt);
		
	} while (evt.eType != appStopEvent);
	
	FrmEraseForm(fp);
	FrmSetActiveForm(NULL);
	FrmDeleteForm(fp);
}

static void uiPrvAppExit(UInt32 osVer)
{
	//in PalmOS3.0+, simply exiting PilotMain is enough to quit an app.
	//before that, to REALLY quit we'll just launch preferences

	if (osVer < sysMakeROMVersion(3,0,0,0,0))
		(void)SysUIAppSwitch(0, DmFindDatabase(0, "Preferences"), sysAppLaunchCmdNormalLaunch, NULL);
}

UInt32 __Startup__(void);
UInt32 __Startup__(void)
{
	void *prevGlobalsP, *globalsP;
	SysAppInfoPtr appInfoP;
	
	if (SysAppStartup(&appInfoP, &prevGlobalsP, &globalsP) == errNone) {
		
		if (appInfoP->cmd == sysAppLaunchCmdNormalLaunch) {
			
			UInt32 osVer;
	
			if (errNone != FtrGet(sysFtrCreator, sysFtrNumROMVersion, &osVer))
				osVer = 0x01000000;
			
			if ((*(volatile UInt8*)(osVer >= sysMakeROMVersion(2,0,0,0,0) ? 0x264 : 0x254)) & 4) {	//warm boot (the global moved in OS2)
				
				if (FrmCustomAlert(1002, "You seem to have warm-booted. Want to reset the baudrate to the default ceiling?", " ", " "))
					uiPrvSetBaudrateLimit(0);
			}
			else {
			
				UInt32 expMgrVer;
				
				if (errNone == FtrGet(sysFileCExpansionMgr, expFtrIDVersion, &expMgrVer) && expMgrVer) {
				
					UInt32 expIter = expIteratorStart;
					UInt16 slotRefNum, slotLibRefNum;
										
					while (expIter != expIteratorStop && errNone == ExpSlotEnumerate(&slotRefNum, &expIter)) {
												
						if (errNone == ExpSlotLibFind(slotRefNum, &slotLibRefNum)) {
						
							UInt8 v1 = 0x80 + (0x7f & SysRandom(0)), v2 = v1;
							UInt16 sz = sizeof(v2);
														
							if (errNone == SlotCustomControl(slotLibRefNum, SASD_API_CREATOR, SASD_VERIFY, &v2, &sz) && sz == sizeof(v2) && (v1 ^ v2) == 0xff) {
								
								ui(slotLibRefNum, slotRefNum);
								goto out;
							}
						}
					}
				}
				
				FrmCustomAlert(1000, "SASD device not found", " " , " ");
			}
			uiPrvAppExit(osVer);
		}
		
out:
		SysAppExit(appInfoP, prevGlobalsP, globalsP);
	}
	
	return 0;
}









